@@ -6,17 +6,18 @@ use BoardStreams::Client::WebSocket 'make_websocket_observable';
 use BoardStreams::Client::Channel;
 use BoardStreams::Client::StructDiff 'patch_state';
 
+use Mojo::JSON 'decode_json', 'encode_json';
 use RxPerl::Mojo ':all';
 use Data::Dump 'dump';
 use List::Util 'min';
 use Carp 'croak';
 
-our $VERSION = "v0.0.7";
+our $VERSION = "v0.0.8";
 
 has 'url';
 has 'connection_status_o';
 
-my @WAIT_TIMES = (0, 1, 3, 5);
+my @WAIT_TIMES = (1, 1, 3, 5);
 
 sub new ($class, $url) {
     my $self = $class->SUPER::new(url => $url);
@@ -61,16 +62,35 @@ sub new ($class, $url) {
     $self->{wses}->subscribe(sub ($x) {
         $self->{ws} = $x;
 
+        my sub pass_along ($data) {
+            my $channel_name = $data->{channel};
+            my $channel_uuid = $data->{channelUUID};
+            if (my $channels_set = $self->{channels}{$channel_name}) {
+                foreach my $channel (values %$channels_set) {
+                    next unless !$channel_uuid or $channel->{uuid} eq $channel_uuid;
+                    $channel->{messages_o}->next($data);
+                }
+            }
+        }
+
+        my %ongoing_messages;
         if ($self->{ws}) {
-            # pass server messages to the channels
-            $self->{ws}->on(json => sub ($, $data) {
-                my $channel_name = $data->{channel};
-                my $channel_uuid = $data->{channelUUID};
-                if (my $channels_set = $self->{channels}{$channel_name}) {
-                    foreach my $channel (values %$channels_set) {
-                        next unless !$channel_uuid or $channel->{uuid} eq $channel_uuid;
-                        $channel->{messages_o}->next($data);
+            $self->{ws}->on(binary => sub ($, $bytes) {
+                $bytes =~ s/^\:([^:]+)\: //s;
+                if (defined(my $bytes_prefix = $1)) {
+                    my ($identifier, $i) = $bytes_prefix =~ /^(.*)\s(\S+)\z/;
+                    if ($i eq '0') {
+                        $ongoing_messages{$identifier} = $bytes;
+                    } elsif ($i =~ /^\d+\z/) {
+                        $ongoing_messages{$identifier} .= $bytes;
+                    } elsif ($i eq 'end') {
+                        $ongoing_messages{$identifier} .= $bytes;
+                        my $data = decode_json(delete $ongoing_messages{$identifier});
+                        pass_along $data;
                     }
+                } else {
+                    my $data = decode_json $bytes;
+                    pass_along $data;
                 }
             });
         }
@@ -83,7 +103,7 @@ sub new ($class, $url) {
 
 sub send ($self, $msg) {
     if ($self->{ws}) {
-        $self->{ws}->send({ json => $msg });
+        $self->{ws}->send({ binary => encode_json $msg });
         # debug
         say "output: ", dump $msg if $ENV{BS_DEBUG};
     } else {
