@@ -9,7 +9,7 @@ use BoardStreams::Util 'string', ':bool';
 use BoardStreams::DBMigrations;
 
 use Mojo::Pg;
-use Mojo::JSON 'to_json', 'from_json', 'encode_json';
+use Mojo::JSON 'to_json', 'from_json', 'encode_json', 'decode_json';
 use Mojo::WebSocket 'WS_PING';
 use Mojo::IOLoop;
 use RxPerl::Mojo ':all';
@@ -18,22 +18,24 @@ use Safe::Isa;
 use Syntax::Keyword::Try;
 use Struct::Diff 'diff';
 use Data::GUID;
-use List::AllUtils qw/ each_array indexes max /;
+use List::AllUtils qw/ each_array indexes max min /;
 use Time::HiRes ();
 use Storable 'dclone';
 use Sys::Hostname; # exports 'hostname' function
 use Carp 'croak';
+use Encode 'encode_utf8';
 
 use experimental 'postderef';
 
-our $VERSION = "v0.0.7";
+our $VERSION = "v0.0.8";
 
 use constant {
-    DEFAULT_HEARTBEAT_INTERVAL       => 5,
-    DEFAULT_HEARTBEAT_TIMEOUT        => 10,
-    DEFAULT_PREFIX                   => 'bs',
-    DEFAULT_PING_INTERVAL            => 15,
+    DEFAULT_HEARTBEAT_INTERVAL        => 5,
+    DEFAULT_HEARTBEAT_TIMEOUT         => 10,
+    DEFAULT_PREFIX                    => 'bs',
+    DEFAULT_PING_INTERVAL             => 15,
     DEFAULT_NOTIFY_PAYLOAD_SIZE_LIMIT => 8000,
+    MAX_WEBSOCKET_SIZE                => 262_144,
 };
 
 my $WORKERS_CHANNEL = '_bs:workers';
@@ -44,7 +46,7 @@ sub register ($self, $app, $config) {
     # config params
     my $heartbeat_interval = $config->{heartbeat_interval} // DEFAULT_HEARTBEAT_INTERVAL;
     my $heartbeat_timeout = $config->{heartbeat_timeout} // DEFAULT_HEARTBEAT_TIMEOUT;
-    my $prefix = $config->{prefix} // DEFAULT_PREFIX;
+    my $bs_prefix = $config->{prefix} // DEFAULT_PREFIX;
     my $ping_interval = $config->{ping_interval} // DEFAULT_PING_INTERVAL;
     my $notify_payload_size_limit = ($config->{notify_payload_size_limit} // DEFAULT_NOTIFY_PAYLOAD_SIZE_LIMIT) - 1;
 
@@ -52,15 +54,15 @@ sub register ($self, $app, $config) {
     my $db_string = $config->{'db_string'} or die "missing db_string configuration option";
     my $pg = Mojo::Pg->new($db_string);
     BoardStreams::DBMigrations->apply_migrations($pg);
-    $app->helper("$prefix.db" => sub { $pg->db });
-    $app->helper("$prefix.pubsub" => sub { $pg->pubsub });
-    $app->helper("$prefix.pubsub_db" => sub { $pg->pubsub->db });
-    my $event_patch_sequence_name = $app->$prefix->db->query("
+    $app->helper("$bs_prefix.db" => sub { $pg->db });
+    $app->helper("$bs_prefix.pubsub" => sub { $pg->pubsub });
+    $app->helper("$bs_prefix.pubsub_db" => sub { $pg->pubsub->db });
+    my $event_patch_sequence_name = $app->$bs_prefix->db->query("
         SELECT pg_get_serial_sequence('event_patch', 'id')
     ")->array->[0];
 
     my sub get_time {
-        return $app->$prefix->db->query("SELECT EXTRACT(EPOCH FROM CURRENT_TIMESTAMP)")
+        return $app->$bs_prefix->db->query("SELECT EXTRACT(EPOCH FROM CURRENT_TIMESTAMP)")
             ->array->[0];
     }
 
@@ -96,7 +98,7 @@ sub register ($self, $app, $config) {
 
         # remove any heartbeat record
         $graceful_stop_o->subscribe(sub {
-            $app->$prefix->lock_state($WORKERS_CHANNEL, sub ($state) {
+            $app->$bs_prefix->lock_state($WORKERS_CHANNEL, sub ($state) {
                 delete $state->{$worker_uuid};
 
                 return undef, $state;
@@ -104,7 +106,7 @@ sub register ($self, $app, $config) {
         });
 
         # register heartbeat
-        $app->$prefix->lock_state($WORKERS_CHANNEL, sub ($state) {
+        $app->$bs_prefix->lock_state($WORKERS_CHANNEL, sub ($state) {
             $last_heartbeat_dt = int get_time;
             $state->{$worker_uuid} = {
                 uuid      => $worker_uuid,
@@ -120,7 +122,7 @@ sub register ($self, $app, $config) {
         )->pipe(
             op_take_until($graceful_stop_o),
         )->subscribe(sub {
-            $app->$prefix->lock_state($WORKERS_CHANNEL, sub ($state) {
+            $app->$bs_prefix->lock_state($WORKERS_CHANNEL, sub ($state) {
                 my $me = $state->{$worker_uuid};
                 if (! $me or $me->{banned}) {
                     Mojo::IOLoop->stop_gracefully;
@@ -166,18 +168,18 @@ sub register ($self, $app, $config) {
         $pg->pubsub->listen('!system' => sub ($pubsub, $payload) {});
     });
 
-    $app->helper("$prefix.add_action", sub ($c, $channel_name, $action_name, $action_sub) {
+    $app->helper("$bs_prefix.add_action", sub ($c, $channel_name, $action_name, $action_sub) {
         BoardStreams::Registry->add_action($channel_name, $action_name, $action_sub);
     });
 
-    $app->helper("$prefix.add_request", sub ($c, $channel_name, $request_name, $request_sub) {
+    $app->helper("$bs_prefix.add_request", sub ($c, $channel_name, $request_name, $request_sub) {
         BoardStreams::Registry->add_request($channel_name, $request_name, $request_sub);
     });
 
-    $app->helper("$prefix.create_channel_or_die", sub ($c, $channel_name, $state = undef, $type = '') {
+    $app->helper("$bs_prefix.create_channel_or_die", sub ($c, $channel_name, $state = undef, $type = '') {
         my $sth;
         try {
-            $sth = $c->$prefix->db->dbh->prepare('INSERT INTO "channel" (name, type, state) VALUES (?, ?, ?)');
+            $sth = $c->$bs_prefix->db->dbh->prepare('INSERT INTO "channel" (name, type, state) VALUES (?, ?, ?)');
             $sth->execute($channel_name, $type, to_json($state));
         } catch {
             my $err = $@;
@@ -190,8 +192,8 @@ sub register ($self, $app, $config) {
         }
     });
 
-    $app->helper("$prefix.create_channel", sub ($c, $channel_name, $state = undef, $type = '') {
-        my $db = $c->$prefix->db;
+    $app->helper("$bs_prefix.create_channel", sub ($c, $channel_name, $state = undef, $type = '') {
+        my $db = $c->$bs_prefix->db;
         ! $db->select('channel',
             ['id'],
             { name => $channel_name },
@@ -206,25 +208,20 @@ sub register ($self, $app, $config) {
         );
     });
 
-    $app->helper("$prefix.lock_state", sub ($c, $channel_names, $sub, %opts) {
+    $app->helper("$bs_prefix.lock_state", sub ($c, $channel_names, $sub, %opts) {
         # opts can be: no_txn, no_ban
         my $multi_mode = length ref $channel_names;
         $channel_names = [$channel_names] if not $multi_mode;
-        my $db = $c->$prefix->db;
+        my $db = $c->$bs_prefix->db;
         my $tx; $tx = $db->begin unless $opts{no_txn};
         my $rows = $db->select('channel',
             [qw/ id name state /],
             { name => {-in => $channel_names} },
             { for => 'update' },
-        )->hashes; # NOTE: is it possible that the row does not exist?
+        )->hashes;
 
-        # if (
-        #     $last_heartbeat_dt and
-        #         time > $last_heartbeat_dt + calculate_hb_expiry_dur() / 2 and
-        #         ! $opts{no_ban}
-        # )
         if (! $opts{no_ban}) {
-            my $workers_state = $c->$prefix->get_state($WORKERS_CHANNEL);
+            my $workers_state = $c->$bs_prefix->get_state($WORKERS_CHANNEL);
             $workers_state->{$worker_uuid}
                 and ! $workers_state->{$worker_uuid}{banned}
                 or die "worker is banned from lock_state";
@@ -330,10 +327,10 @@ sub register ($self, $app, $config) {
                 if (length $notification > $notify_payload_size_limit) {
                     my $i = 0;
                     while (my $len = length $notification) {
-                        my $prefix =
+                        my $notification_prefix =
                             $len > $notify_payload_size_limit ? $i++ :
                             'end';
-                        substr($notification, 0, 0, ":$prefix: ");
+                        substr($notification, 0, 0, ":$notification_prefix: ");
 
                         $db->notify($channel_name, substr($notification, 0, $notify_payload_size_limit, ''));
                     }
@@ -350,19 +347,46 @@ sub register ($self, $app, $config) {
         return $multi_mode ? \@new_states : $new_states[0];
     });
 
-    $app->helper("$prefix.on_join", sub ($c, $channel_name, $join_sub) {
+    $app->helper("$bs_prefix.on_join", sub ($c, $channel_name, $join_sub) {
         BoardStreams::Registry->add_join($channel_name, $join_sub);
     });
 
-    $app->helper("$prefix.on_leave", sub ($c, $channel_name, $leave_sub) {
+    $app->helper("$bs_prefix.on_leave", sub ($c, $channel_name, $leave_sub) {
         BoardStreams::Registry->add_leave($channel_name, $leave_sub);
     });
 
-    $app->helper("$prefix.on_cleanup", sub ($c, $channel_name, $cleanup_sub) {
+    $app->helper("$bs_prefix.on_cleanup", sub ($c, $channel_name, $cleanup_sub) {
         BoardStreams::Registry->add_cleanup($channel_name, $cleanup_sub);
     });
 
-    $app->helper("$prefix.initialize_client", sub ($self) {
+    # NOTE: turn this sub into async later
+    $app->helper("$bs_prefix.send", sub ($c, $data, $identifier) {
+        my sub get_max_size {
+            return min($c->tx->max_websocket_size, MAX_WEBSOCKET_SIZE);
+        }
+
+        my $bytes = encode_json $data;
+        if (length $bytes < get_max_size) {
+            $c->send({binary => $bytes});
+            return;
+        }
+
+        my $i = 0;
+        $identifier =~ s/\:/{}/g;
+        while (length $bytes) {
+            my $max_size = get_max_size;
+            my $bytes_prefix =
+                length $bytes > $max_size ? $i++ :
+                'end';
+            $bytes_prefix = ":$identifier $bytes_prefix: ";
+
+            substr $bytes, 0, 0, encode_utf8($bytes_prefix);
+            my $piece = substr $bytes, 0, $max_size, '';
+            $c->send({binary => $piece});
+        }
+    });
+
+    $app->helper("$bs_prefix.initialize_client", sub ($self) {
         # reject connection, if pubsub is not active
         return $self->rendered(500) if not $pubsub_connected;
 
@@ -381,7 +405,8 @@ sub register ($self, $app, $config) {
         });
 
         # trap incoming messages
-        $self->on(json => async sub ($c, $hash) {
+        $self->on(binary => async sub ($c, $bytes) {
+            my $hash = decode_json($bytes);
 
             # join
             if ($hash->{type} eq 'join') {
@@ -402,7 +427,7 @@ sub register ($self, $app, $config) {
 
                 my $was_added = BoardStreams::Registry->add_pair($c, $channel_name);
                 if ($was_added) {
-                    my $o = get_listener_observable($c, $channel_name, $prefix);
+                    my $o = get_listener_observable($c, $channel_name, $bs_prefix);
                     my $s = $o->subscribe();
                     $c->stash->{subscriptions}{channels}{$channel_name} = $s;
                 }
@@ -410,7 +435,7 @@ sub register ($self, $app, $config) {
                 # NOTE: in the async case in the future, make sure the subscription is still there
                 # after the query, before sending state to user
 
-                my $db = $c->$prefix->db;
+                my $db = $c->$bs_prefix->db;
                 my ($channel_id, $event_id, $state) = $db->select('channel',
                     [qw/ id event_id state /],
                     { name => $channel_name },
@@ -435,7 +460,7 @@ sub register ($self, $app, $config) {
                     )->hashes->reverse;
 
                     foreach my $event_row (@$latest_event_rows) {
-                        $c->send({json => {
+                        $c->$bs_prefix->send({
                             type        => 'event_patch',
                             channel     => $channel_name,
                             channelUUID => $channel_uuid,
@@ -444,17 +469,17 @@ sub register ($self, $app, $config) {
                             data        => {
                                 event => from_json($event_row->{event}),
                             },
-                        }});
+                        }, "event_patch-$channel_name");
                     }
                 }
 
-                $c->send({json => {
+                $c->$bs_prefix->send({
                     type        => 'state',
                     channel     => $channel_name,
                     channelUUID => $channel_uuid,
                     id          => int $event_id,
                     data        => from_json($state),
-                }});
+                }, "state-$channel_name");
             }
 
             # leave
@@ -499,21 +524,21 @@ sub register ($self, $app, $config) {
                 }
 
                 if ($ok) {
-                    $c->send({json => {
+                    $c->$bs_prefix->send({
                         type      => 'response',
                         channel   => $channel_name,
                         requestId => $request_id,
                         result    => $return_value,
                         duration  => $duration,
-                    }});
+                    }, "response-$channel_name-$request_id");
                 } else {
-                    $c->send({json => {
+                    $c->$bs_prefix->send({
                         type      => 'response',
                         channel   => $channel_name,
                         requestId => $request_id,
                         error     => $err,
                         duration  => $duration,
-                    }});
+                    }, "response-$channel_name-$request_id");
                 }
             }
         });
@@ -536,24 +561,24 @@ sub register ($self, $app, $config) {
         });
     });
 
-    $app->helper("$prefix.joined", sub ($c, $channel_name) {
+    $app->helper("$bs_prefix.joined", sub ($c, $channel_name) {
         return BoardStreams::Registry->has_pair($c, $channel_name);
     });
 
-    $app->helper("$prefix.worker_uuid", sub ($c) {
+    $app->helper("$bs_prefix.worker_uuid", sub ($c) {
         return $worker_uuid;
     });
 
-    $app->helper("$prefix.get_state", sub ($c, $channel_name) {
-        my $db = $c->$prefix->db;
+    $app->helper("$bs_prefix.get_state", sub ($c, $channel_name) {
+        my $db = $c->$bs_prefix->db;
         my $state = from_json $db->select('channel',
             ['state'],
             { name => $channel_name },
         )->hash->{state};
     });
 
-    $app->helper("$prefix.alive_workers", sub ($c) {
-        my $workers_state = $c->$prefix->get_state($WORKERS_CHANNEL);
+    $app->helper("$bs_prefix.alive_workers", sub ($c) {
+        my $workers_state = $c->$bs_prefix->get_state($WORKERS_CHANNEL);
         return {
             map {( $_, 1 )}
                 grep ! $workers_state->{$_}{banned},
@@ -561,15 +586,15 @@ sub register ($self, $app, $config) {
         };
     });
 
-    $app->helper("$prefix.global_cleanup", sub ($c) {
-        my $workers_state = $c->$prefix->get_state($WORKERS_CHANNEL);
+    $app->helper("$bs_prefix.global_cleanup", sub ($c) {
+        my $workers_state = $c->$bs_prefix->get_state($WORKERS_CHANNEL);
         my $time = get_time;
         foreach my $worker (values %$workers_state) {
             if ($time > $worker->{heartbeat} + calculate_hb_expiry_dur()) {
-                my $db = $c->$prefix->db;
+                my $db = $c->$bs_prefix->db;
 
                 # 0. set 'banned' property of worker to true
-                $c->$prefix->lock_state($WORKERS_CHANNEL, sub ($state) {
+                $c->$bs_prefix->lock_state($WORKERS_CHANNEL, sub ($state) {
                     $state->{$worker->{uuid}}{banned} = true
                         if exists $state->{$worker->{uuid}};
                     return undef, $state;
@@ -600,8 +625,8 @@ sub register ($self, $app, $config) {
                     # call cleanup sub
                     my $channel_name = $channel->{name};
                     if (my $cleanup_sub = BoardStreams::Registry->get_cleanup($channel_name)) {
-                        $c->$prefix->lock_state($channel_name, sub ($state) {
-                            my $alive_workers = $c->$prefix->alive_workers;
+                        $c->$bs_prefix->lock_state($channel_name, sub ($state) {
+                            my $alive_workers = $c->$bs_prefix->alive_workers;
                             my $new_state = $cleanup_sub->($channel_name, $state, $alive_workers);
 
                             return undef, $new_state;
@@ -626,7 +651,7 @@ sub register ($self, $app, $config) {
                 }
 
                 # 3. delete worker from :heartbeats
-                $c->$prefix->lock_state($WORKERS_CHANNEL, sub ($state) {
+                $c->$bs_prefix->lock_state($WORKERS_CHANNEL, sub ($state) {
                     delete $state->{$worker->{uuid}};
                     return undef, $state;
                 }, no_ban => 1);
@@ -639,8 +664,8 @@ sub register ($self, $app, $config) {
 
     # system channels
     {
-        $app->$prefix->create_channel($WORKERS_CHANNEL, {});
-        $app->$prefix->global_cleanup();
+        $app->$bs_prefix->create_channel($WORKERS_CHANNEL, {});
+        $app->$bs_prefix->global_cleanup();
     }
 }
 
